/**
 * Date:            Time:       User:               Description:
 * 18.04.2006       17:36       Smart Max           creation of the file Block.java
 *
 */
package org.max.trie.impl;

import org.max.trie.INode;
import org.max.trie.ISymbol;
import org.max.trie.IWritableNode;
import org.max.util.IndexedSet;

import java.io.Serializable;
import java.util.*;

/**
 * Block
 *
 * @author  Maxim Podkolzine
 */
class Block implements Serializable {
    private static final long serialVersionUID = 482993088763784L;

    // Data

    /**
     * Holds the name of block.
     * <p>
     * Using this string blocks will seek for far children
     */
    private String myName;

    /**
     * Holds the set of nodes in sub-trie of this block
     */
    private IndexedSet<INode> myNodes;

    /**
     * Holds the list of internal edges of this block
     * (i.e. the edges of sub-trie)
     */
    private List<Edge> myEdges;

    /**
     * Holds the list of far edges of this block
     * (i.e. the edges that refer to nodes in other block)
     */
    private List<FarLink> myFarLinks;

    /**
     * Class represents an edge: holds indices of its end (<tt>from</tt> and <tt>to</tt>).
     * It is serializable and uses default java serialization.
     * <p>
     * Note that the symbol on edges is not stored: it can be found using incoming symbol of child node
     * {@link Node#myIncomingSymbol}, and both nodes are serialized in one block.
     */
    private static class Edge implements Serializable {
        private static final long serialVersionUID = 788510583038969589L;

        private int from;
        private int to;

        public Edge(int from, int to) {
            this.from = from;
            this.to = to;
        }
    }

    /**
     * Class represents the far link (an edge between two nodes in separate blocks).
     * It holds the index of node, from which link starts (<tt>from</tt>), the symbol on it (<tt>symbol</tt>)
     * and the name of destination block (<tt>blockName</tt>). It is the root of corresponding block that
     * child node of an edge.
     * Class is serializable and uses default java serialization.
     */
    private static class FarLink implements Serializable {
        private static final long serialVersionUID = 5458984101178031100L;

        private int from;
        private ISymbol symbol;
        private String blockName;

        public FarLink(int from, ISymbol symbol, String blockName) {
            this.from = from;
            this.symbol = symbol;
            this.blockName = blockName;
        }
    }




    // Constructor

    /**
     * Creates a new block with specified name
     *
     * @param   name                                the block name
     */
    public Block(String name) {
        myName = name;
        myNodes = new IndexedSet<INode>();
        myEdges = new ArrayList<Edge>();
        myFarLinks = new ArrayList<FarLink>();
    }




    // Main interface

    /**
     * Returns the block name
     *
     * @return                                      the block name
     */
    public String getName() {
        return myName;
    }

    /**
     * Returns the size of block. It is measured in number of nodes in it.
     *
     * @return                                      the block size
     */
    public int size() {
        return myNodes.size();
    }

    // Construct methods

    /**
     * Adds a separate node to the block.
     * <p>
     * The nodes can not duplicate. That's why node is added only if it did not exist in block before.
     *
     * @param   node                                the node to be added
     */
    public void addNode(INode node) {
        myNodes.add(node);
    }

    /**
     * Adds two nodes, connected by an edge.
     * <p>
     * The nodes can not duplicate. That's why nodes are added only if they did not exist in block before.
     * In contrast, the edge between them is added anyway. Thus, each edge of the sub-trie must be added
     * only once.
     *
     * @param   node                                the parent node
     * @param   child                               the child node
     */
    public void addEdge(INode node, INode child) {
        int from = myNodes.add(node);
        int to = myNodes.add(child);
        myEdges.add(new Edge(from, to));
    }

    /**
     * Adds the node and far link on child in another block.
     * <p>
     * The nodes can not duplicate. That's why node is added only if it did not exist in block before.
     * In contrast, the edge between nodes is added anyway. Thus, each edge of the sub-trie must be added
     * only once.
     *
     * @param   node                                the parent node
     * @param   symbol                              the symbol on edge
     * @param   blockName                           the name of far block
     */
    public void addEdge(INode node, ISymbol symbol, String blockName) {
        int from = myNodes.add(node);
        myFarLinks.add(new FarLink(from, symbol, blockName));
    }

    /**
     * Adds another block to this block.
     * <p>
     * In fact, method just unites the sets of nodes (new nodes obtain new indices and existed nodes preserve
     * old ones) and adds all edges from new block (all: internal and far).
     * <p>
     * Note, that two blocks (this and prepared to be added) can have common nodes, but can not have common
     * edges (in this case, they would be duplicated).
     *
     * @param   block                               the block to be added
     */
    @SuppressWarnings({"UnnecessaryLocalVariable"})
    public void addBlock(Block block) {
        int[] newIndices = new int[block.size()];

        for (INode node : block.myNodes) {
            int previous = block.myNodes.getIndex(node);
            int current = this.myNodes.add(node);
            newIndices[previous] = current;
        }

        for (Edge edge : block.myEdges) {
            edge.from = newIndices[edge.from];
            edge.to = newIndices[edge.to];
            this.myEdges.add(edge);
        }

        for (FarLink farLink : block.myFarLinks) {
            farLink.from = newIndices[farLink.from];
            this.myFarLinks.add(farLink);
        }
    }

    // Reconstruct method

    /**
     * Returns the root of the corresponding sub-trie.
     * <p>
     * Method constructs the trie structure by data in existence (by set of nodes and edges).
     * Method automatically checks that the data is correct and non-conflicting and can throw
     * an exception in case of troubles.
     * <p>
     * All nodes, except root and leafs, will have full information in trie. The leafs will have
     * far children instead of direct children (the references to another blocks). And the root node
     * will have no parent. The parent can be determined only when this block will be uploaded from
     * another block (see {@link Node#getChild(org.max.trie.ISymbol)}).
     * <p>
     * Method is expected to be used in the following way:
     * <ul>
     * <li> first the block is constructed by dividing the trie
     * <li> after this block is serialized (note that each block is serialized separately)
     * <li> when needed the block is deserialized
     * <li> and finally this method constructs the needed sub-trie
     * <ul>
     *
     * @return                                      the root of sub-trie
     */
    @SuppressWarnings({"unchecked"})
    public IWritableNode getSubTrie() {
        /*
        At first we determine the root of subtrie.
        This must be the node with no incoming edges.
        */

        assert(myNodes.size() - myEdges.size() == 1);
        boolean[] inner = new boolean[myNodes.size()];

        for (Edge edge : myEdges) {
            assert(!inner[edge.to]);
            inner[edge.to] = true;
        }

        int root = -1;
        for (int i = 0; i < inner.length; i++) {
            if (!inner[i]) {
                root = i;
                break;
            }
        }
        assert(root >= 0);

        /*
        Create the graph: for each node we must have the list of children:
        close children (i.e. connected by edge) and far children (i.e. connected by far link).
        */

        ArrayList<Integer>[] closeChildren = new ArrayList[myNodes.size()];
        ArrayList<FarLink>[] farChildren = new ArrayList[myNodes.size()];

        for (int i = 0; i < myNodes.size(); i++) {
            closeChildren[i] = new ArrayList<Integer>();
            farChildren[i] = new ArrayList<FarLink>();
        }

        for (Edge edge : myEdges) {
            closeChildren[edge.from].add(edge.to);
        }

        for (FarLink link : myFarLinks) {
            farChildren[link.from].add(link);
        }

        /*
        After this we traverse all nodes and fill parents and outcoming edges (nodes or far block names).
        */

        Queue<Integer> queue = new LinkedList<Integer>();
        queue.add(root);

        while (!queue.isEmpty()) {
            int v = queue.poll();
            Node node = (Node) myNodes.get(v);
            node.createChildrenMap();

            for (Integer childIndex : closeChildren[v]) {
                IWritableNode child = (IWritableNode) myNodes.get(childIndex);
                child.setParent(node);
                node.addChild(child);

                queue.add(childIndex);
            }

            for (FarLink link : farChildren[v]) {
                node.addChild(link.symbol, link.blockName);
            }
        }

        return (IWritableNode) myNodes.get(root);
    }
}
