/**
 * Date:            Time:       User:               Description:
 * 14.03.2006       20:20       Smart Max           creation of the file PatriciaTrie.java
 *
 */
package org.max.trie.impl;

import org.max.trie.IWritablePatriciaTrie;
import org.max.trie.ISymbol;
import org.max.trie.IWritableNode;
import org.max.trie.INode;

import java.util.*;

/**
 * PatriciaTrie
 *
 * @author  Maxim Podkolzine
 */
class PatriciaTrie implements IWritablePatriciaTrie {
    // Data

    /**
     * Holds the root node of trie
     */
    private IWritableNode myRoot;

    /**
     * Holds the map of full symbols sequences on edges.
     * <p>
     * Since, each node (except root) have exactly one incoming edge,
     * the map correctly confront edges with true list of symbols on it.
     * <p>
     * Note, that without this map a trie have only one symbol on each edge (the starting one).
     * Thus, searching in a trie does not guarantee that the obtained node would be the desired.
     */
    private Map<IWritableNode, List<ISymbol>> myMap;

    /**
     * Holds whether the trie is modifiable
     */
    private boolean myModifiable;




    // Constructor

    /**
     * Creates an empty Patricia trie, which has only root node.
     * <p>
     * After this, a trie is modifiable and supports adding of strings
     * (see {@link #add(java.util.List<org.max.trie.ISymbol>)})
     */
    public PatriciaTrie() {
        myRoot = new Node();
        myMap = new HashMap<IWritableNode, List<ISymbol>>();
        myModifiable = true;
    }

    /**
     * Creates an Patricia trie with specified root node, map of symbols on edges and the modification value.
     * <p>
     * This constructor is expected to be used for deserialization of a trie
     * (see methods of class {@link PatriciaTrieDeserializer}).
     *
     * @param   root                                the root of trie
     * @param   map                                 the map of symbols on edges
     * @param   modifiable                          the modification value
     */
    public PatriciaTrie(IWritableNode root, Map<IWritableNode, List<ISymbol>> map, boolean modifiable) {
        myRoot = root;
        myMap = map;
        myModifiable = modifiable;
    }




    // Main interface

    // Add method

    /**
     * Creates the exact copy of specified list in another list.
     * <p>
     * Method creates a new list of <tt>ISymbol</tt>s and adds all elements starting with
     * <tt>fromIndex</tt> and up to <tt>toIndex - 1</tt>. Thus, the size of result list will
     * be exactly <tt>toIndex - fromIndex</tt>.
     *
     * @param   list                                the source list
     * @param   fromIndex                           the starting index
     * @param   toIndex                             the ending index
     * @return                                      the list of <tt>ISymbol</tt>s, containing
     *                                              the requested elements of source list
     */
    private List<ISymbol> copyList(List<ISymbol> list, int fromIndex, int toIndex) {
        List<ISymbol> result = new ArrayList<ISymbol>(toIndex - fromIndex);

        for (int i = fromIndex; i < toIndex; i++) {
            result.add(list.get(i));
        }

        return result;
    }

    /**
     * Holds the constant list, which indicates that corresponding edge contains only one symbol.
     * <p>
     * In detail, each edge of Patricia trie has the list of symbols. But due to lossy compression,
     * only one symbol is stored in trie structure, and other symbols can be stored separately in
     * special map (see {@link #myMap}). Note, that this map is needed only for modifiable tries,
     * because while retrieving in a trie only one symbol on each edge is used.
     * <p>
     * Since first symbol on edge can be easily found, the list of symbols is actual only if
     * the corresponding edge has several (more than one) symbols. And other edges can have some
     * default list, which indicates that edge contains exactly one symbol. For this purpose,
     * the constant <tt>DefaultList</tt> has introduces.
     */
    private final List<ISymbol> DefaultList = new ArrayList<ISymbol>();

    /**
     * Puts the list of symbols on the incoming edge of specified node.
     * <p>
     * Method simply puts the list to the map <tt>myMap</tt>, except for the case when the list
     * contains only one element. In this case, the {@link #DefaultList} is added.
     *
     * @param   node                                the node, before which the list must be added
     * @param   list                                the list of symbols, which are needed to be added
     */
    private void putList(IWritableNode node, List<ISymbol> list) {
        assert(list.size() > 0);

        if (list.size() == 1) {
            myMap.put(node, DefaultList);
        } else {
            myMap.put(node, list);
        }
    }

    /**
     * Adds the specified string of <tt>ISymbol<tt>s to the Patricia trie.
     * <p>
     * If the trie is not modifiable (see {@link #myModifiable}), the <tt>UnsupportedOperationException</tt>
     * raises.
     * <p>
     * While adding the whole list of symbols on each edge is used (not only first one). So it is
     * possible that incoming string will divide the existing edge on two edges by new node.
     * In any case, at most two new nodes will be added to the trie independently from length
     * of string and size of trie.
     *
     * @param   string                              the list of <tt>ISymbol</tt>s to be added
     * @return                                      the newly created node, which corresponds to
     *                                              the given string in this trie
     * @throws  UnsupportedOperationException       if the trie if unmodifiable
     */
    public IWritableNode add(List<ISymbol> string) {
        if (!myModifiable) {
            throw new UnsupportedOperationException("Patricia trie is unmodifiable");
        }

        IWritableNode node = myRoot;
        int depth;

        while (true) {
            depth = node.getDepth();
            assert(depth <= string.size());

            if (depth == string.size()) {
                return node;
            }

            ISymbol symbol = string.get(depth);
            IWritableNode child = node.getChild(symbol);

            if (child == null) {
                break;
            }

            List<ISymbol> list = myMap.get(child);
            if (list == DefaultList) {
                node = child;
                continue;
            }
            assert((list != null) && (list.size() > 0) && (symbol.equals(list.get(0))));

            int i = 1;
            while (i < list.size() && i + depth < string.size() && string.get(i + depth).equals(list.get(i))) {
                i++;
            }

            if (i == list.size()) {
                node = child;
            } else {
                IWritableNode intermediate = new Node(depth + i, node, symbol);
                node.changeChild(symbol, intermediate);
                child.setParent(intermediate);
                child.setIncomingSymbol(list.get(i));
                intermediate.addChild(child);
                                                                  // Note:
                putList(intermediate, copyList(list, 0, i));      // here must be used copyList(list, ...) instead of
                putList(child, copyList(list, i, list.size()));   // list.subList(...) because of serialization.
                                                                  // Point is that class java.util.AbstractList.SubList
                if (i + depth == string.size()) {                 // is not Serializable.
                    return intermediate;
                }

                node = intermediate;
                depth += i;
                break;
            }
        }

        IWritableNode newNode = new Node(string.size(), node, string.get(depth));
        node.addChild(newNode);
                                                                      // Note:
        List<ISymbol> list = copyList(string, depth, string.size());  // here must be used copyList(string, ...)
        putList(newNode, list);                                       // instead of string.subList(...) because the
                                                                      // string can be modifiable and its external
        return newNode;                                               // modification must not tell upon a trie.
    }

    // Get method

    /**
     * Returns the node of a trie, which corresponds to specified string of symbols in this trie.
     * <p>
     * Note, that the map {@link #myMap} is not available in this method (by definition of Patricia trie).
     * So the symbols in string are compared with taking into account the depth of node and the first symbol
     * on edge only.
     * <p>
     * For example, consider the trie that was constructed by strings: "<tt>abcdef</tt>" and "<tt>abcxyzw</tt>".
     * This trie has the root, one inner node and two leafs with corresponding depths: <tt>0</tt>, <tt>3</tt>,
     * <tt>6</tt> and <tt>7</tt>. The real symbols on edges must be "<tt>abc</tt>", "<tt>def</tt>" and
     * "<tt>xyzw</tt>", but due to strong compression the trie will hold only "<tt>a</tt>", "<tt>d</tt>" and
     * "<tt>x</tt>". So any of the following requests will return positive result: "<tt>a??d??</tt>" and
     * "<tt>a??x???</tt>".
     * <p>
     * Consequently, this method (and even the trie idea itself) does not guarantee that found data matches
     * the search key, so additional verification is needed. But, it does guarantee that in case of negative
     * result the key is not contained in a trie. In that case, method returns <tt>null</tt>.
     *
     * @param   string                              the requested string of symbols
     * @return                                      the node corresponding to the string or <tt>null</tt> if
     *                                              trie does not contain such string.
     */
    public INode get(List<ISymbol> string) {
        INode node = myRoot;
        while (true) {
            int depth = node.getDepth();

            if (depth == string.size()) {
                return node;
            } else if (depth > string.size()) {
                return null;
            }

            INode n = node.getChild(string.get(depth));
            if (n != null) {
                node = n;
            } else {
                return null;
            }
        }
    }

    // Other IWritablePatriciaTrie methods

    /**
     * Returns the root of the trie
     *
     * @return                                      the root of the trie
     */
    public INode getRoot() {
        return myRoot;
    }

    /**
     * Returns the size of the trie (i.e. the number of nodes in it)
     *
     * @return                                      the trie size
     */
    public int size() {
        return myMap.size() + 1;
    }

    /**
     * Returns whether the trie is modifiable
     *
     * @return                                      true iff the trie is modifiable
     */
    public boolean isModifiable() {
        return myModifiable;
    }

    /**
     * Sets the modification value of the trie.
     * <p>
     * Note, that to make the trie modifiable, the map of symbols on edges must be setted.
     *
     * @param   modifiable                          the new modification value
     */
    public void setModifiable(boolean modifiable) {
        myModifiable = modifiable;
    }

    /**
     * Releases the map of full symbols sequences on edges.
     * <p>
     * More precisely, the method simply nulls the map:
     * <pre>
     *  myMap = null;
     * </pre>
     */
    public void release() {
        myMap = null;
    }

    /**
     * Returns the map of full symbols sequences on edges.
     * <p>
     * Method does not represented in any public interface and can be used by package-local classes only.
     * The main purpose of this method is serialization of the trie without losing of volatility.
     *
     * @return                                      the map of symbols on edges
     */
    public Map<IWritableNode, List<ISymbol>> getMap() {
        return myMap;
    }

    /**
     * Sets the map of full symbols sequences on edges.
     * <p>
     * Method does not represented in any public interface and can be used by package-local classes only.
     * The main purpose of this method is serialization of the trie without losing of volatility.
     * It is not expected to use this method to null the map (see {@link #release()}).
     *
     * @param   map                                 the map of symbols on edges
     */
    public void setMap(Map<IWritableNode, List<ISymbol>> map) {
        myMap = map;
    }
}
