/**
 * Date:            Time:       User:               Description:
 * 19.04.2006       11:58       Smart Max           creation of the file PatriciaTrieDivider.java
 *
 */
package org.max.trie.impl;

import static org.max.NativeSystem.*;

import org.max.trie.IPatriciaTrie;
import org.max.trie.INode;
import org.max.trie.ISymbol;

import java.util.List;
import java.util.ArrayList;
import java.util.Collections;

/**
 * PatriciaTrieDivider
 *
 * @author  Maxim Podkolzine
 */
class PatriciaTrieDivider {
    // Main interface

    /**
     * Holds the maximal size of block
     */
    private static int ourBlockSize = 10;

    /**
     * Returns the maximal block size
     *
     * @return                                      the maximal block size
     */
    public static int getBlockSize() {
        return ourBlockSize;
    }

    /**
     * Sets the maximal block size
     *
     * @param   blockSize                           new maximal block size
     */
    public static void setBlockSize(int blockSize) {
        if (blockSize > 0) {
            PatriciaTrieDivider.ourBlockSize = blockSize;
        }
    }

    // Divide method

    /**
     * Holds the root directory containing blocks.
     * <p>
     * It is used to indicate absolute name of interblock references.
     */
    private String myDirectory;

    /**
     * Holds current block index.
     * <p>
     * All blocks will be named in series: "1.block", "2.block", ...
     */
    private int myBlock;

    /**
     * Creates and returns new block
     *
     * @return                                      new block
     */
    private Block newBlock() {
        return new Block((myBlock++) + ".block");
    }

    /**
     * Holds the list of blocks, which divide the trie.
     * <p>
     * During the dividing algorithm, some blocks will unite and considered as one.
     * After dividing all result blocks will be collected in this list.
     */
    private List<Block> myResultBlocks;

    /**
     * Returns the full size of sub-trie (measured in nodes number).
     * <p>
     * The sub-trie is counted as a whole (without assumption that some nodes are already
     * put into blocks)
     *
     * @param   node                                the root node of sub-trie
     * @return                                      the sub-trie size
     */
    private int getSubTrieSize(INode node) {
        int result = 1;

        for (ISymbol symbol : node.getOutcomingSymbols()) {
            INode child = node.getChild(symbol);
            result += getSubTrieSize(child);
        }

        return result;
    }

    /**
     * Puts the whole sub-trie to the specified block.
     * <p>
     * The size of sub-trie must fit for putting into one block.
     *
     * @param   node                                the root node of sub-trie
     * @param   block                               the block
     */
    private void putSubTrie(INode node, Block block) {
        block.addNode(node);

        for (ISymbol symbol : node.getOutcomingSymbols()) {
            INode child = node.getChild(symbol);
            block.addEdge(node, child);
            putSubTrie(child, block);
        }

        assert(block.size() <= ourBlockSize);
    }

    /**
     * Proceeds a sub-trie: divides it on blocks and returns the root block.
     * <p>
     * This is the main method of dividing algorithm. It lies in recursive proceeding of sub-tries.
     * The processing of current sub-trie includes:
     * <ul>
     * <li> getting the sizes of all sub-tries (before dividing)
     * <li> recursion: all large sub-tries (which size is more than <tt>ourBlockSize</tt>)
     *      are divided recursively; after this the sizes combine into array <tt>s[1..n]</tt>:
     *      it contains the full sizes of small sub-tries and the sizes of root blocks
     *      that divide large sub-tries
     * <li> all proper sub-tries (i.e. small enough) are put in separate blocks;
     *      after whis step there would be the array of block sizes <tt>s[1]..s[n]</tt>,
     *      each <tt>1&lt;=s[i]&lt;=ourBlockSize</tt>, and <tt>s[0]==1</tt> (root)
     * <li> the least blocks and root block are united in one block; this block and other not-united
     *      blocks form the result division
     * </ul>
     *
     * @param rootNode
     * @return                                      the root block
     */
    private Block proceedSubTrie(INode rootNode) {
        int childrenNumber = rootNode.getOutcomingSymbols().size();
        int[] sizes = new int[childrenNumber];
        INode[] nodes = new Node[childrenNumber];
        Block[] blocks = new Block[childrenNumber];

        int i = 0;
        Block result = null;
        int min = ourBlockSize;
        int minI = -1;
        for (ISymbol symbol : rootNode.getOutcomingSymbols()) {
            nodes[i] = rootNode.getChild(symbol);

            if (getSubTrieSize(nodes[i]) > ourBlockSize) {      // the sub-trie is too large, must be divided
                blocks[i] = proceedSubTrie(nodes[i]);
            } else {                                            // the sub-trie is appropriate to add as a whole
                blocks[i] = newBlock();
                putSubTrie(nodes[i], blocks[i]);
            }

            sizes[i] = blocks[i].size();
            if (sizes[i] < min) {
                result = blocks[i];
                min = sizes[i];
                minI = i;
            }

            i++;
        }

        boolean[] used = new boolean[childrenNumber];

        if (result == null) {
            result = newBlock();
            result.addNode(rootNode);
        } else {
            result.addEdge(rootNode, nodes[minI]);

            used[minI] = true;

            while (true) {
                min = ourBlockSize;
                minI = -1;
                for (i = 0; i < sizes.length; i++) {
                    if (!used[i] && sizes[i] < min) {
                        min = sizes[i];
                        minI = i;
                    }
                }

                if (minI == -1 || result.size() + min > ourBlockSize) {
                    break;
                }

                used[minI] = true;
                result.addEdge(rootNode, nodes[minI]);
                result.addBlock(blocks[minI]);
            }
        }

        // for all not-resulting blocks we add far edge to the root node;
        // also we verify the blocks sizes:
        // it must be guaranteed that size of each block <= ourBlockSize
        assert(result.size() <= ourBlockSize);

        for (i = 0; i < childrenNumber; i++) {
            if (!used[i]) {
                assert(blocks[i].size() <= ourBlockSize);
                myResultBlocks.add(blocks[i]);

                result.addEdge(rootNode, nodes[i].getIncomingSymbol(), myDirectory + blocks[i].getName());
            }
        }

        return result;
    }

    /**
     * Divides the trie on blocks.
     * <p>
     * A trie is divided on sub-tries, i.e. each block contains the connected part of trie,
     * and all the parts are binded by interblock references. The maximal admissible block
     * size is equal to {@link #ourBlockSize}; method tries to divide the Patricia trie
     * as optimal as possible.
     * <p>
     * The result is list of blocks, the last one is root block.
     *
     * @param   trie                                the trie needed to be divided
     * @param   directory                           the destination directory
     * @return                                      the blocks, which divide the trie
     */
    public List<Block> divide(IPatriciaTrie trie, String directory) {
        myDirectory = directory + Separator;
        myBlock = 0;
        myResultBlocks = new ArrayList<Block>();

        Block rootBlock = proceedSubTrie(trie.getRoot());
        myResultBlocks.add(rootBlock);

        return Collections.unmodifiableList(myResultBlocks);
    }
}
