/**
 * Date:            Time:       User:               Description:
 * 16.03.2006       12:24       Smart Max           creation of the file BidirectionalMap.java
 * 22.04.2006       11:35       Smart Max           map can be modifiable and unmodifiable
 *
 */
package org.max.util;

import java.util.Map;
import java.util.HashMap;
import java.io.Serializable;

/**
 * BidirectionalMap
 *
 * @author  Maxim Podkolzine
 */
public class BidirectionalMap<A,B> implements Serializable {
    private static final long serialVersionUID = 982450134816399834L;

    // Data

    /**
     * Holds whether the map is modifiable
     */
    private boolean myModifiable;

    /**
     * Holds the direct map
     */
    private Map<A,B> myDirectMap;

    /**
     * Holds the reverse map
     */
    private Map<B,A> myReverseMap;




    // Constructor

    /**
     * Creates a new bidirectional map.
     * <p>
     * Initially, it is empty (the size is <tt>0</tt>) and modifiable.
     */
    public BidirectionalMap() {
        myModifiable = true;
        myDirectMap = new HashMap<A,B>();
        myReverseMap = new HashMap<B,A>();
    }




    // Main interface

    /**
     * Returns the size of the map
     *
     * @return                                      the size of map
     */
    public int size() {
        return myDirectMap.size();
    }

    /**
     * Puts a new pair of elements to the map.
     * <p>
     * After adding the element <tt>a</tt> can be found by <tt>b</tt> (see {@link  #getA(Object)}) and
     * element <tt>b</tt> can be found by <tt>a</tt> (see {@link  #getB(Object)}).
     * <p>
     * To peform this operation the map must be modifiable (see {@link #isModifiable()}).
     *
     * @param   a                                   the element from A
     * @param   b                                   the element from B
     * @throws  UnsupportedOperationException       if the map is unmodifiable
     * @throws  IllegalArgumentException            if one of elements <tt>a</tt> or <tt>b</tt> is
     *                                              already contained in corresponding set
     * @see     #containsA(Object)
     * @see     #getA(Object)
     * @see     #containsB(Object)
     * @see     #getB(Object)
     * @see     #isModifiable()
     */
    public void put(A a, B b) {
        if (!myModifiable) {
            throw new UnsupportedOperationException("The bidirectional map is unmodifiable");
        }

        if (containsA(a) || containsB(b)) {
            throw new IllegalArgumentException("Pair (" + a + ", " + b + ") can not be added");
        }

        myDirectMap.put(a, b);
        myReverseMap.put(b, a);
    }

    /**
     * Returns whether element <tt>a</tt> is contained in domain of map
     *
     * @param   a                                   the element from A
     * @return                                      true if <tt>a</tt> exists in domain of map
     * @see     #containsB(Object)
     */
    public boolean containsA(A a) {
        return myDirectMap.containsKey(a);
    }

    /**
     * Returns whether element <tt>b</tt> is contained in image of map
     *
     * @param   b                                   the element from B
     * @return                                      true if <tt>b</tt> exists in image of map
     * @see     #containsA(Object)
     */
    public boolean containsB(B b) {
        return myReverseMap.containsKey(b);
    }

    /**
     * Returns the value to which this map maps the specified key <tt>b</tt>, or
     * <tt>null</tt> if the map contains no mapping for this key
     *
     * @param   b                                   the element from B
     * @return                                      the correspong element from <tt>A</tt> or
     *                                              <tt>null</tt> if map contains no mapping for <tt>a</tt>
     * @see     #getB(Object)
     */
    public A getA(B b) {
        return myReverseMap.get(b);
    }

    /**
     * Returns the value to which this map maps the specified key <tt>a</tt>, or
     * <tt>null</tt> if the map contains no mapping for this key
     *
     * @param   a                                   the element from A
     * @return                                      the correspong element from <tt>B</tt> or
     *                                              <tt>null</tt> if map contains no mapping for <tt>a</tt>
     * @see     #getA(Object)
     */
    public B getB(A a) {
        return myDirectMap.get(a);
    }

    /**
     * Returns a string representation of the map
     *
     * @return                                      the string representation of the map
     */
    public String toString() {
        return myDirectMap.toString();
    }

    /**
     * Returns whether the map is modifiable
     *
     * @return                                      true iff the map is modifiable
     * @see     #setModifiable(boolean)
     */
    public boolean isModifiable() {
        return myModifiable;
    }

    /**
     * Sets the modification value of map
     *
     * @param   modifiable                          the new modification value
     * @see     #isModifiable()
     */
    public void setModifiable(boolean modifiable) {
        myModifiable = modifiable;
    }
}
