/**
 * Date:            Time:       User:               Description:
 * 07.05.2006       14:10:08       Smart Max           creation of the file PatriciaTrieDeserializerTest.java
 *
 */
package org.max.trie.impl;

import static org.max.NativeSystem.*;
import static org.max.trie.impl.TestUtil.*;

import junit.framework.TestCase;

import org.max.trie.IPatriciaTrie;
import org.max.trie.INode;
import org.max.trie.IWritablePatriciaTrie;
import org.max.xml.index.XMLBuilder;
import org.max.RandomXML;

import java.io.IOException;

import net.sf.xanswer.indexing.IndexException;

/**
 * PatriciaTrieDeserializerTest
 */
public class PatriciaTrieDeserializerTest extends TestCase {
    public void test_Simple() throws IOException, ClassNotFoundException, IndexException {
        XMLBuilder xmlBuilder = new XMLBuilder();
        XMLBuilder.restoreDefaults();
        xmlBuilder.addXMLDocument("xml" + Separator + "doc1.xml");
        IPatriciaTrie trie = xmlBuilder.getIndexFabric().getTrie();

        String location = "test" + Separator + "doc1 [4]";
        createDirectory(location);
        PatriciaTrieSerializer serializer = new PatriciaTrieSerializer(trie, location);
        serializer.writeTrie(4);

        PatriciaTrieDeserializer deserializer = new PatriciaTrieDeserializer();
        IPatriciaTrie trie2 = deserializer.readTrie(location);

        assertEqualSubTries(trie.getRoot(), trie2.getRoot());
    }

    private void checkTrie(String ... names) throws IOException, ClassNotFoundException {
        XMLBuilder builder = new XMLBuilder();
        for (String name : names) {
            try {
                builder.addXMLDocument("xml" + Separator + name);
            } catch (IndexException e) {
                e.printStackTrace();
                fail();
            }
        }

        IPatriciaTrie trie = builder.getIndexFabric().getTrie();

        StringBuilder locationBuilder = new StringBuilder("test").append(Separator);
        for (String name : names) {
            if (name.endsWith(".xml")) {
                locationBuilder.append(name, 0, name.length() - 4);
            } else {
                locationBuilder.append(name);
            }
            locationBuilder.append(' ');
        }
        String location = locationBuilder.deleteCharAt(locationBuilder.length() - 1).toString();

        //System.out.println(location + ":\n" + builder + "\n");

        PatriciaTrieSerializer serializer = new PatriciaTrieSerializer(trie, location);

        createDirectory(location);
        serializer.writeTrie();

        PatriciaTrieDeserializer deserializer = new PatriciaTrieDeserializer();
        IPatriciaTrie trie2 = deserializer.readTrie(location);

        INode root1 = trie.getRoot();
        INode root2 = trie2.getRoot();
        assertEqualSubTries(root1, root2);
    }

    public void test_Unicode() throws IOException, ClassNotFoundException {
        checkTrie("unicode.xml");
    }

    public void test_Mscorlib() throws IOException, ClassNotFoundException {
        checkTrie("mscorlib.xml");
    }

    public void test_Colors_Web() throws IOException, ClassNotFoundException {
        checkTrie("colors.xml", "web.xml");
    }

    public void test_SystemDrawing_SystemWindowsForms() throws IOException, ClassNotFoundException {
        checkTrie("System.Drawing.xml", "System.Windows.Forms.xml");
    }

    // Random tries

    static {
        createDirectory("test" + Separator + "random");
    }

    public void test_Random_Medium() throws IOException, ClassNotFoundException {
        RandomXML random = new RandomXML();
        String name = "random" + Separator + System.currentTimeMillis() + ".xml";
        random.createXML("xml" + Separator + name, 10000, 5);
        checkTrie(name);
    }

    public void test_Random_Large() throws IOException, ClassNotFoundException {
        RandomXML random = new RandomXML();
        String name = "random" + Separator + System.currentTimeMillis() + ".xml";
        random.createXML("xml" + Separator + name, 100000, 2000, 4, 100, 0.2, 6, 0.4, 10);
        XMLBuilder.setAddPathPrefixes(false);
        checkTrie(name);
    }

    /**
     * Note: to run this test: rename it in test_Random_ExtremelyLarge() and
     * uncomment output in checkTrie(String ... names)
     *
     * This method was runned with such statistics:
     * java -Xmx500M -ea
     *  Total time of work: 1'772,765 sec (~29,5 min)
     *  Total memory usage: 400'407 Kb
     *  Result:             passed
     *
     * That was the output:
     *  test\random\1147086191328:
     *  Index Fabric builder of XML
     *  trie size = 851094
     *  text map size = 64
     *  attr map size = 199809
     *  tag map size = 10000
     *  total map size = 209873
     *
     * File 1147086191328.xml size ~ 28 Mb
     * Serialized it keeps 75'701'104 bytes in 18'192 files
     */
    public void _test_Random_ExtremelyLarge() throws IOException, ClassNotFoundException {
        RandomXML random = new RandomXML();
        String name = "random" + Separator + System.currentTimeMillis() + ".xml";
        random.createXML("xml" + Separator + name, 1000000, 10000, 5, 200, 0.2, 6, 0.4, 20);
        XMLBuilder.setAddPathPrefixes(false);
        checkTrie(name);
    }

    // Full serialization

    public void test_FullSerialization1() throws IOException, ClassNotFoundException, IndexException {
        XMLBuilder xmlBuilder = new XMLBuilder();
        XMLBuilder.restoreDefaults();
        xmlBuilder.addXMLDocument("xml" + Separator + "doc1.xml");
        xmlBuilder.addXMLDocument("xml" + Separator + "doc2.xml");
        IPatriciaTrie trie = xmlBuilder.getIndexFabric().getTrie();

        String location = "test" + Separator + "full-1 [4]";

        PatriciaTrieSerializer serializer = new PatriciaTrieSerializer(trie, location);

        createDirectory(location);
        serializer.writeFullTrie(4);

        PatriciaTrieDeserializer deserializer = new PatriciaTrieDeserializer();
        IWritablePatriciaTrie trie2 = deserializer.readFullTrie(location);

        INode root1 = trie.getRoot();
        INode root2 = trie2.getRoot();
        assertEqualSubTries(root1, root2);

        xmlBuilder = new XMLBuilder();
        xmlBuilder.addXMLDocument("xml" + Separator + "text.xml");
        IPatriciaTrie trie3 = xmlBuilder.getIndexFabric().getTrie();
        addTrie(trie2, trie3);
    }
}
