/**
 * Date:            Time:       User:               Description:
 * 18.04.2006       17:41:03       Smart Max           creation of the file PatriciaTrieTest.java
 *
 */
package org.max.trie.impl;

import junit.framework.TestCase;
import org.max.trie.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * PatriciaTrieTest
 */
public class PatriciaTrieTest extends TestCase {
    private static class TestSymbol implements ISymbol {
        // Data

        private String s;

        // Constructor

        public TestSymbol(String s) {
            this.s = s;
        }

        // Main interface

        public boolean equals(Object o) {
            if (!(o instanceof PatriciaTrieTest.TestSymbol)) {
                return false;
            }

            final TestSymbol that = (TestSymbol) o;

            return (s != null ? s.equals(that.s) : that.s == null);
        }

        public int hashCode() {
            return (s != null ? s.hashCode() : 0);
        }

        public String toString() {
            return s;
        }

        public static List<ISymbol> getList(String path) {
            List<ISymbol> result = new ArrayList<ISymbol>();

            for (int j = 0; j < path.length(); j++) {
                result.add(new TestSymbol(String.valueOf(path.charAt(j))));
            }

            return result;
        }
    }

    /**
     * A simple test with adding few symbols to trie.
     */
    public void test_Simple() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        ISymbol a = new TestSymbol("a");
        ISymbol b = new TestSymbol("b");
        ISymbol c = new TestSymbol("c");
        trie.add(Arrays.asList(a));
        trie.add(Arrays.asList(a, b));
        trie.add(Arrays.asList(a, b, c));

        assertNotNull(trie.get(Arrays.asList(a)));
        assertNotNull(trie.get(Arrays.asList(a, b)));
        assertNotNull(trie.get(Arrays.asList(a, b, c)));
    }

    private void check_ConsecutiveAdding(String ... paths) {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        for (String path : paths) {
            for (int i = 1; i <= path.length(); i++) {
                trie.add(TestSymbol.getList(path.substring(0, i)));
            }
        }

        for (String path : paths) {
            List<ISymbol> key = TestSymbol.getList(path);
            assertNotNull(trie.get(key));
        }
    }

    /**
     * Test on consecutive adding: to add a string s[1..n] we request all strings
     * s[1..1], s[1..2], ..., s[1..n].
     * No edge of trie will have more than 1 label.
     */
    public void test_ConsecutiveAdding() {
        check_ConsecutiveAdding("abc", "abd");
        check_ConsecutiveAdding("abc", "abc", "abd", "abd");
        check_ConsecutiveAdding("abc", "acd", "abcd");
        check_ConsecutiveAdding("abcd", "abcc", "abacc", "abdc", "abda");
        check_ConsecutiveAdding("aaa", "bbb", "ccc", "ddd", "ab", "ca", "bc");
    }

    private void checkLeaf(INode node, int depth) {
        assertNotNull(node);
        assertEquals(depth, node.getDepth());
        assertEquals(0, node.getOutcomingSymbols().size());
    }

    /**
     * Test adds strings in such order that new nodes must divide an existing edge.
     * For example, we add string "prefix" first, and after this string "prefi".
     * After first operation a trie has one edge with labels "prefix", but
     * the second request adds new node on it; the trie will have edges "prefi" and "x".
     */
    public void test_AddPrefixes1() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        trie.add(TestSymbol.getList("prefix"));
        trie.add(TestSymbol.getList("prefi"));
        trie.add(TestSymbol.getList("pref"));
        trie.add(TestSymbol.getList("pre"));
        trie.add(TestSymbol.getList("pr"));
        trie.add(TestSymbol.getList("p"));

        assertNotNull(trie.get(TestSymbol.getList("p")));
        assertNotNull(trie.get(TestSymbol.getList("pr")));
        assertNotNull(trie.get(TestSymbol.getList("pre")));
        assertNotNull(trie.get(TestSymbol.getList("pref")));
        assertNotNull(trie.get(TestSymbol.getList("prefi")));
        assertNotNull(trie.get(TestSymbol.getList("prefix")));
        assertNull(trie.get(TestSymbol.getList("prefix1")));
    }

    /**
     * Test adds strings in such order that new nodes must divide an existing edge.
     */
    public void test_AddPrefixes2() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        trie.add(TestSymbol.getList("Patricia"));
        trie.add(TestSymbol.getList("Pat"));
        trie.add(TestSymbol.getList("Patric"));
        trie.add(TestSymbol.getList("P"));
        trie.add(TestSymbol.getList("Patr"));

        assertNotNull(trie.get(TestSymbol.getList("P")));
        assertNull(trie.get(TestSymbol.getList("Pa")));
        assertNotNull(trie.get(TestSymbol.getList("Pat")));
        assertNotNull(trie.get(TestSymbol.getList("Patr")));
        assertNull(trie.get(TestSymbol.getList("Patri")));
        assertNotNull(trie.get(TestSymbol.getList("Patric")));
        assertNull(trie.get(TestSymbol.getList("Patrici")));
        assertNotNull(trie.get(TestSymbol.getList("Patricia")));
        assertNull(trie.get(TestSymbol.getList("Patricia1")));
    }

    /**
     * Test constructs and seeks in sample tree.
     */
    public void test_SmallTree() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        trie.add(TestSymbol.getList("cat"));
        trie.add(TestSymbol.getList("castle"));
        trie.add(TestSymbol.getList("cash"));
        trie.add(TestSymbol.getList("casting"));
        trie.add(TestSymbol.getList("castiron"));
        trie.add(TestSymbol.getList("far"));
        trie.add(TestSymbol.getList("fast"));

        checkLeaf(trie.get(TestSymbol.getList("cat")), 3);
        checkLeaf(trie.get(TestSymbol.getList("castle")), 6);
        checkLeaf(trie.get(TestSymbol.getList("cash")), 4);
        checkLeaf(trie.get(TestSymbol.getList("casting")), 7);
        checkLeaf(trie.get(TestSymbol.getList("castiron")), 8);
        checkLeaf(trie.get(TestSymbol.getList("far")), 3);
        checkLeaf(trie.get(TestSymbol.getList("fast")), 4);
    }

    /**
     * Test constructs and seeks in a tree of medium size.
     * For this it adds all substrings of string "abcdefghabcd" (length=12).
     * This test adds strings in order of its length increasing.
     */
    public void test_MediumTree_IncreaseLength() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        String s = "abcdefghabcd";
        for (int i = 0; i < s.length(); i++) {
            for (int j = i; j < s.length(); j++) {
                trie.add(TestSymbol.getList(s.substring(i, j)));
            }
        }

        for (int i = 0; i < s.length(); i++) {
            for (int j = i; j < s.length(); j++) {
                assertNotNull(trie.get(TestSymbol.getList(s.substring(i, j))));
            }
        }
    }

    /**
     * Test constructs and seeks in a tree of medium size.
     * For this it adds all substrings of string "abcdefghabcd" (length=12).
     * This test adds strings in order of its length decreasing.
     */
    public void test_MediumTree_DecreaseLength() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        String s = "abcdefghabcd";
        for (int i = 0; i < s.length(); i++) {
            for (int j = s.length(); j >= i; j--) {
                trie.add(TestSymbol.getList(s.substring(i, j)));
            }
        }

        for (int i = 0; i < s.length(); i++) {
            for (int j = i; j < s.length(); j++) {
                assertNotNull(trie.get(TestSymbol.getList(s.substring(i, j))));
            }
        }
    }

    /**
     * Memory test.
     * After introduction of DefaultList in PatriciaTrie memory usage reduced with such statistics:
     * before (delta usage): 4950-5100 Kb
     * after  (delta usage): 4350-4600 Kb
     */
    public void test_MemoryOptimization() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < 300; i++) {
            if ((i*i) % 59 < 30) {
                builder.append('a');
            } else {
                builder.append('b');
            }
        }

        for (int i = 0; i < builder.length(); i++) {
            for (int j = i; j < builder.length(); j++) {
                trie.add(TestSymbol.getList(builder.substring(i, j)));
            }
        }

        for (int i = 0; i < builder.length(); i++) {
            for (int j = i; j < builder.length(); j++) {
                assertNotNull(trie.get(TestSymbol.getList(builder.substring(i, j))));
            }
        }
    }

    /**
     * Test on path modification after request.
     * While adding to trie, some lists are copied in map.
     * It is important to make a new list with such elements
     * (not to use method sublist method of path)
     */
    public void test_RemoveListElements1() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        List<ISymbol> list1 = PatriciaTrieTest.TestSymbol.getList("castle");
        trie.add(list1);
        list1.clear();
        checkLeaf(trie.get(TestSymbol.getList("castle")), 6);

        List<ISymbol> list2 = TestSymbol.getList("casting");
        trie.add(list2);
        list2.clear();

        checkLeaf(trie.get(TestSymbol.getList("castle")), 6);
        checkLeaf(trie.get(TestSymbol.getList("casting")), 7);
    }

    /**
     * Another test on path modification
     */
    public void test_RemoveListElements2() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        List<ISymbol> list1 = PatriciaTrieTest.TestSymbol.getList("abcdef");
        trie.add(list1);

        List<ISymbol> list2 = TestSymbol.getList("abc");
        trie.add(list2);

        List<ISymbol> list3 = TestSymbol.getList("abcxyz");
        trie.add(list3);

        list1.clear();
        list2.clear();
        list3.clear();

        assertNotNull(trie.get(TestSymbol.getList("abc")));
        checkLeaf(trie.get(TestSymbol.getList("abcdef")), 6);
        checkLeaf(trie.get(TestSymbol.getList("abcxyz")), 6);
    }

    public void test_LossyCompression() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        trie.add(TestSymbol.getList("abcdef"));
        trie.add(TestSymbol.getList("abcxyzw"));
        
        checkLeaf(trie.get(TestSymbol.getList("a??d??")), 6);
        checkLeaf(trie.get(TestSymbol.getList("a??x???")), 7);
    }

    /**
     * Test checks the paths that can not be found in trie.
     */
    public void test_NotFoundPaths() {
        IWritablePatriciaTrie trie = PatriciaTrieFactory.getInstance().newWritablePatriciaTrie();

        trie.add(TestSymbol.getList("cat"));
        trie.add(TestSymbol.getList("catling"));
        trie.add(TestSymbol.getList("collection"));
        trie.add(TestSymbol.getList("dog"));
        trie.add(TestSymbol.getList("dolly"));

        assertNull(trie.get(TestSymbol.getList("a")));
        assertNull(trie.get(TestSymbol.getList("ca")));
        assertNull(trie.get(TestSymbol.getList("cats")));
        assertNull(trie.get(TestSymbol.getList("catlin")));
        assertNull(trie.get(TestSymbol.getList("core")));
        assertNull(trie.get(TestSymbol.getList("collage")));
        assertNull(trie.get(TestSymbol.getList("cell")));
        assertNull(trie.get(TestSymbol.getList("d")));
        assertNull(trie.get(TestSymbol.getList("dogs")));
        assertNull(trie.get(TestSymbol.getList("dol")));
        assertNull(trie.get(TestSymbol.getList("dollly")));
        assertNull(trie.get(TestSymbol.getList("dope")));

        // Note, that:
        // trie.get(TestSymbol.getList("catrine")) != null
        // trie.get(TestSymbol.getList("dolby")) != null
        // The reason: the lossy compression
    }
}
